@@ -471,6 +471,9 @@ const char* meta_event_detail_to_string (int d);
 void meta_display_queue_retheme_all_windows (MetaDisplay *display);
 void meta_display_retheme_all (void);
 
+void meta_display_set_cursor_theme (const char *theme, 
+				    int         size);
+
 void     meta_display_ping_window              (MetaDisplay        *display,
 						MetaWindow         *window,
 						Time                timestamp,
